//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svg_exception.h"
#include "svg_req_filter.h"
#include "svg_filter.h"
#include "svg_filter_primitive.h"
#include "svg_fe_alpha_mask.h"
namespace svg
{

	//Constructor
	//The svg::root element should assign the SVG graph attrrs and elems
	//storages! Then it will come back to the sender.
	//Arguments:
	//	fltr:	the filter definition
	//	elm:	element, wich has to have this filter applied
	//	gtrans:	global transformation
	//	bounds:	actual clip box
	//	pixels:	traget pixel format (result buffer)
	//	extent_ratio: optional. determines, how bigger will be the intermediate
	//				  buffers comparing to the target area (some extent is required
	//				  for example for the gaussian blurs - otherwise they would be 
	//				  distored at the edges)
	req_filter::req_filter(svg::filter &fltr,shape &elm, transform gtrans, 
								rectangle bounds, pixfmt &pixels, double extent_ratio): 
		request(request::filter),
		m_filter(fltr),
		m_x(0), m_y(0), m_width(0), m_height(0),
		m_elm(elm),
		m_trans(gtrans),
		m_pixels(pixels)
	{
		//assign the element and attribute storage to this request:
		req_prepare prepare(request::prepare,*this);
		elm.to_root(prepare);

		//obtain the filter region in the relative version (expressed in the
		//relative units of selected m_elem bounding box):
		rectangle rgon;
		
		rectangle box = fltr.region();
		
		if(fltr.relative_units() != 1) //when filterUnits = userSpaceOnUse,
									   //we have to translate it into RELATIVE
								       //units for this particular m_elem
		{
			rectangle rbox = elm.bounding_box();
			double wdth = rbox.x2 - rbox.x1;
			double hght = rbox.y2 - rbox.y1;
			
			//rgon is calculated as relative to actual bounding box
			rgon = rectangle( (box.x1 - rbox.x1)/wdth,
							  (box.y1 - rbox.y1)/hght,
							  (box.x2 - rbox.x1)/wdth,
							  (box.y2 - rbox.y1)/hght
							);
		}
		else
		{
			rgon = box; //for relative units, filter box is the relative filter region
		}

		//Here is the trick (maybe not exactly as in the SVG spec): we will apply
		//the RELATIVE coordinates of the filter region to the bounding box of m_elem
		//measured IN THE ABSOLUTE PIXEL COORDINATES. This will produce the ABSOLUTE
		//pixel coordinates of the filter region. According SVG spec, we should 
		//apply filter to the nominal coordinates, and then transfrom the result 
		//into absolute pixel coordinates. It would take much more time, and 
		//also loose some accuracy.

		transform trans = m_elm.trans(); //trans is the transformation from element
		trans *= m_trans;				 //nominal coordinates (i.e. its bounding box)
										 //to pixel coordinates of the result buffer

		box = elm.bounding_box(trans); //now it's extents of the m_elem in absolute 
									   //pixel coordinates of the final 
									   //rendering
		trans.reset();
		trans.scale(box.x2 - box.x1, box.y2 - box.y1);
		trans.translate(box.x1, box.y1);
		trans.transform(&(rgon.x1), &(rgon.y1));
		trans.transform(&(rgon.x2), &(rgon.y2));

		//move the origin for the eventual rendering to upper left of 
		//the filter region (in this form it can be applied to obtain the source
		//graphic in the intermediate buffer:
		m_trans.translate(-rgon.x1, -rgon.y1);

		//clip the region to not exceed too much the result buffer: 
		//(no more than ::extent_ratio scale)
		{
			double x0 = (bounds.x1 + bounds.x2)*0.5;
			double y0 = (bounds.y1 + bounds.y2)*0.5;
			double dx = (bounds.x2 - bounds.x1)*extent_ratio*0.5;
			double dy = (bounds.y2 - bounds.y1)*extent_ratio*0.5;

			rectangle clip(x0-dx, y0-dy, x0+dx, y0+dy);

			//when whole rgon is NOT inside clip:
			if(!is_inside(rgon,clip))
			{
				//here: box is the common part of two areas:
				box = agg::intersect_rectangles(rgon, clip);
				m_trans.translate(rgon.x1-box.x1, rgon.y1 - box.y1);
				rgon = box;
			}
			
		}

		//Now we can round the result and store it into instance fields:
		m_x = agg::iround(rgon.x1); 
		m_y = agg::iround(rgon.y1); 
		m_width = agg::iround(rgon.x2 - rgon.x1); if (m_width < 0) m_width = 0;
		m_height = agg::iround(rgon.y2 - rgon.y1); if (m_height < 0) m_height = 0;

		//and calculate the transformation for the filter primitives:
		if (fltr.relative_content_units() == 1)
		{
			m_prm_trans  = trans; // notice above, that trans contains actually 
		}						  // the transformation to absolute nominal units
		else
		{
			m_prm_trans = transform(); //just identity
		}
		m_prm_trans *= elm.trans(); //from nominal coords to object transformed coords...
		m_prm_trans *= m_trans;		//...and to the pixel coords of the result buffer

	}	
//----------------------------------------------------------------------------
// Classes - different kinds of the buffers:
	class plain_buffer: public req_filter::buffer
	{
		public:
			//Creates a plain buffer, filled with transparent black
			//Arguments
			//	width:		width, in pixels
			//	height:		height, in pixels
			plain_buffer(unsigned width, unsigned height)
			{
				unsigned size = width * height * pixfmt::pix_width;
				m_picture = new agg::int8u [size];
				memset(m_picture, 0, size); //fill with transparent black: all zeros!

				m_rbuf.attach(m_picture, width, height, width* pixfmt::pix_width);
				m_pixels.attach(m_rbuf);
			}

			virtual ~plain_buffer()
			{
				if(m_picture)
				{
					delete [] m_picture;
					m_picture = NULL;
				}
			}

		protected:
			agg::int8u *m_picture;
	};//plain_buffer
//----------------------------------------------------------------------------
// Implementation of SourceGraphic
	class source_graphic_buffer: public plain_buffer
	{
		public:
			//Creates a source buffer, filled with transparent black
			//Arguments
			//	fltr:		this filter object
			//	elm:		element, which will be fltered
			//	gtrans:		global transfrom
			//	width:		width, in pixels
			//	height:		height, in pixels
			source_graphic_buffer(svg::filter &fltr, element &elm, transform gtrans, 
												unsigned width, unsigned height):
			plain_buffer(width, height)
			{
				pixfmt pf(m_rbuf);

				rectangle clip(0,0,pf.width(),pf.height());
				req_render rnd(pf,clip,rgba8(0,0,0,0));
				req_prepare prepare(request::prepare_render, rnd);

				elm.to_root(prepare);
				if(rnd.can_query())
				{
					//final preparation of the render request: 
					//restrictions:
					rnd.restricted_filter = fltr.id();

					//when we are rednering the filter (a clip_path):
					//no strokes!
					if (elm.is_of("clipPath")) rnd.set(req_render::no_stroke);

					rnd.trans = gtrans;
					elm.on_request(rnd);
				}
			}

	};//source_graphic_buffer
//----------------------------------------------------------------------------
// TO DO: implementation of FillPaint, StrokePaint buffers.
// Similar to implementation of SourceGraphic above, but, after creation the 
// req_render, there will be a little trick:
// 1. a simple rectangle, that matches the size of m_pixels, is prepared
// as the vertex source. Use the agg::poly_plain_adaptor<int> for this purpose,
// wrapped with a simple class (to be written) that implemements the vsource 
// interface around it.
// This rectangle is the vertex source. You have to prepare that it will 
// produce the rectangle that matches the m_pixels buffer size.
// 2. Simplify the rendering style: read the effective style of the element,
// and set the Stroke or Fill style in a helper styleset variable to proper one.
// Initialize the painter, to let them prepare the internal transform!
// 3. Fill the m_pixels with single call to req_render::perform(vsource, transform, lstyle). 
//----------------------------------------------------------------------------
// Used for implementation of BackgroundImage as a reference to resulted pixfmt
	class ref_buffer: public req_filter::buffer
	{
		public:
			//Arguments
			//	source:		the source buffer
			//	x:			distance from source's left edge to left edge of
			//				the referenced area, in pixels
			//	y:			distance from source's top edge to top edge of
			//				the referenced area, in pixels
			//	width:		width, of the referenced area, in pixels
			//	height:		height, of the referenced area, in pixels
			ref_buffer(pixfmt &source, unsigned x, unsigned y, 
											unsigned width, unsigned height)
			{
				m_pixels.attach(source, x, y, x+width-1, y+width -1);
			}

	};//ref_buffer
//----------------------------------------------------------------------------
// Implementation for SourceAlpha, BackgroundAlpha
	class alpha_buffer: public plain_buffer
	{
		public:
			//Arguments
			//	source:		the source buffer (full color)
			//	x:			distance from source's left edge to left edge of
			//				the referenced area, in pixels
			//	y:			distance from source's top edge to top edge of
			//				the referenced area, in pixels
			//	width:		width, of the referenced area, in pixels
			//	height:		height, of the referenced area, in pixels
			alpha_buffer(pixfmt_pre &source, unsigned x, unsigned y, 
										     unsigned width, unsigned height):
			plain_buffer(width, height)
			{
				rgba8 color;
				unsigned dx, dy;
				for(dy = 0; dy < height; dy++)
				{
					for(dx = 0; dx < width; dx++)
					{
						color = source.pixel(x+dx,y+dy);
						if (color.a > 0) //m_pixels is already filled with
						{				 //transparent black - see plain_buffer
							color.r = color.g = color.b = 0; //leave alpha, only
							m_pixels.copy_pixel(dx,dy,color);
						}
					}
				}
			}

	};//alpha_buffer
//----------------------------------------------------------------------------
	//Adds a new intermediate buffer, having given name
	//Arguments:
	//	name:	name of the new picture buffer
	//			Predefined SVG names, as:
	//				"SourceGraphic"
	//				"SourceAlpha"
	//				"BackgroundImage"
	//				"BackgroundAlpha"
	//				"FillPaint"
	//				"StrokePaint"
	//			are recognized and treated according SVG spec
	req_filter::reference &req_filter::add(const char *name)
	{
		req_filter::buffer *buf = NULL;
		//----------------- predefined SVG buffers --------------------
		// "" is the result buffer - initially filled with source graphic...
		if (strcmp(name,"SourceGraphic") == 0 || strcmp(name, "") == 0) 
		{
			buf = new source_graphic_buffer(m_filter, m_elm, m_trans, m_width, m_height);
		}

		if (strcmp(name,"BackgroundImage") == 0) 
		{
			buf = new ref_buffer(m_pixels, m_x, m_y, m_width, m_height);
		}

		if (strcmp(name,"SourceAlpha") == 0) 
		{
			buf = new alpha_buffer(get("SourceGraphic"), m_x, m_y, m_width, m_height);
		}

		if (strcmp(name,"BackgroundAlpha") == 0) 
		{
			buf = new alpha_buffer(get("BackgroundImage"), m_x, m_y, m_width, m_height);
		}
		//FillPaint and StrokePaint will be implemented when rendering of a rectangle
		//is available
		// "" is the result buffer - initially filled with source graphic...
		//-----------  buffers expected by clip_path and mask ---------
		if (strcmp(name,"#clip") == 0 || strcmp(name, "#mask") == 0) 
		{
			//render its own childern into the input buffer - required by fe_mask 
			//or fe_clip filter primitives
			buf = new source_graphic_buffer(m_filter, m_filter, 
														m_prm_trans, m_width, m_height);
		}
		//-------------------------------------------------------------

		if(!buf) //when name is no predefined name: create another plain buffer
		{
			buf = new plain_buffer(width(),height());
		}
		
		int pos = m_buffers.size();

		if (buf)
		{
			m_buffers.add(reference());
			m_buffers[pos].assign(name, buf); 
		}
		else
		{
			char msg[128];
			sprintf(msg,"req_filter::add: itermediate buffer for : '%s' has not been created.", name);
			throw new exception(msg);
		}
		
		return m_buffers[pos];
	}//end add

	//Returns the reference struct. having given name. If the name does not
	//exists, yet - it creates it as a plain buffer, and then returns!
	//Arguments:
	//	name:	name of the buffer (local to particular filter element)
	req_filter::reference &req_filter::select(const char *name)
	{
		unsigned i;
		reference *result = NULL;

		//Usually there will be just few buffers - there is no need
		//to use any advanced search algorithm:
		for (i = 0; !result && i < m_buffers.size(); i++)
		{
			if(m_buffers[i].is(name)) result = &(m_buffers[i]);
		}

		if (!result) result = &(add(name)); 

		if (!result)
		{
           throw exception("filters: Cannot prepare an image buffer named '%s'", name);
		}
		else 
		{
			return *result;
		}
	}
} // namespace svg
