//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_TSPAN_INCLUDED
#include "svg_shape.h"
#include "svg_req_render.h"
#include "svg_linear_text_position.h"
#include "svg_earray.h"
#define SVG_TSPAN_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// tspan: a text content in the SVG graph
//////////////////////////////////////////////////////////////////////
class tspan : public shape
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		tspan(element *parent, unsigned level): shape(parent, level), 
		m_text(m_content) //normally the m_text references the default content
						  //(to be overriden in the child class - svg::tref)
		{					
			m_fcount = 0;
			m_x = NULL;
			m_y = NULL;
			m_dx = NULL;
			m_dy = NULL;
			m_rotation = NULL;
		}

		//Static constructor - used by the svg::text to create and store a new 
		//svg::tspan element to handle its direct content
		//Arguments:
		//	prs:		parser;
		//	parent:		the text element, that will be the parent of this element
		//	level:		the level of nesting in the source XML tree (just 
		//				parent->level()+1 here)
		//	utf8:		the initial text content. It can be appended later
		//	len:		length od the utf8 text
		//	commited:	true, when the created element has to be already commited
		static element *create(parser &prs, element *parent, unsigned level, 
										const char *utf8, unsigned len = 0, bool commited = true)
		{
			tspan *elm = new tspan(parent, level);
			if (elm) 
			{
				prs.store(*elm,"",false,false); //store unnamed tspan
				elm->append(utf8, len);
				if (commited) elm->commit(prs);
			}
			return elm;
		}

		virtual const char * tag() {return "tspan";};

		virtual void on_content(parser &prs, const char* s, int len) 
		{
			append(s, (unsigned) len);
		}


		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (strcmp(name, "x") == 0) 
			{
				m_x = create(prs,value);
			}
			else 
			if (strcmp(name, "y") == 0) 
			{
				m_y = create(prs,value);
			}
			else 
			if (strcmp(name, "dx") == 0) 
			{
				m_dx = create(prs,value);
			}
			else 
			if (strcmp(name, "dy") == 0) 
			{
				m_dy = create(prs,value);
			}
			else 
			if (strcmp(name, "rotate") == 0) 
			{
				m_rotation = create(prs,value);
			}
			else
			{
				shape::on_attribute(prs,name,value);
			}

		}//on_attribute

	public:

		//Group of helper functions, to mark in an easy way the access to the
		//array attributes:
		double x(request &req, unsigned index, double deflt) 
		{
			return get_value(req, m_x, index, deflt);
		}

		double y(request &req, unsigned index, double deflt) 
		{
			return get_value(req, m_y, index, deflt);
		}

		double dx(request &req, unsigned index, double deflt) 
		{
			return get_value(req, m_dx, index, deflt);
		}

		double dy(request &req, unsigned index, double deflt) 
		{
			return get_value(req, m_dy, index, deflt);
		}
		
		//rotation: returns a value converted to radians (SVG keeps rotations in degrees)
		double rot(request &req, unsigned index, double deflt) 
		{
			return agg::deg2rad(get_value(req, m_rotation, index, agg::rad2deg(deflt)));
		}

		//Appends another fragment of a text to the text already accumulated
		//Arguments:
		//	utf8:	text content, (encoded as utf-8) - may not be null-terminated
		//	len:	number of characters to analyze and convert into unicode
		void append(const char *utf8, unsigned len);

		//Simplified version of the append method: for null-terminated strings
		//Arguments:
		//	utf8:	text content, (encoded as utf-8)
		void append(const char *utf8) {append(utf8, strlen(utf8));}

		//Returns number of characters tored in the tspan:
		unsigned characters() {return m_text.size();}

		//this element is NEVER rendered directly:
	    virtual handle render(request &req)
		{return skip(req);}

		//Renders or measures the content of a single <tspan>. When this instance 
		//contains its own x,y atributes, which enables it to fix the text position, 
		//or the text position is already set up by a previous tspan, the text will
		//be measured/rendered.
		//Arguments:
		//	req:	a svg::req_render or svg::req_measure request
		//	tpos:	current text position (may be reset inside this method)
		//Returns the id of the next element
		handle render(req_geometry &req, text_position &tpos);		

		//Return text, contained in this element
		//Helper function, used by tref to obtain the content of the referenced text
		string &content() {return m_content;}
	protected:
		//Helper function, used in the x(), y(), dx(), dy(), rot() methods
		//Arguments:
		//	req:	actual request object
		//	arr:	the array, which should supply the value
		//	index:	index of the value within the array (may be outside its bounds!)
		//	deflt:	default value, when the index is outside the bounds of the array
		//			(or the array is NULL)
		//Returns a double value
		double get_value(request &req, earray *arr, unsigned index, double deflt)
		{
			if(arr)		return arr->get(req,index,deflt);
				else	return deflt;
		}

		//creates an one of m_x, m_y, m_dx, m_dy, m_rotation elements, and 
		//updates the internal state (created just to not duplicate identical
		//code in the on_attribute() method.
		//Arguments
		//	prs:	parser
		//	spec:	specification of the array xml attribute: a sequence of 
		//			numbers, separated by spaces or commas
		//Returns a pointer to created and stored element
		earray *create(parser &prs,const char *spec)
		{
			earray *arr = (earray *) earray::create(prs, this, level()+1,spec);
			if(arr)
			{
				if (m_fcount < arr->count()) 
				{
					m_fcount = arr->count();
				}
				prs.store(*arr,"",false,false);
				arr->commit(prs);
			}
			return arr;
		}//create

	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		string	m_content; //sequence of the unicode characters, to be drawn
		string	&m_text;   //this reference is prepared for <tref> element,
						   //which use not its own content. Normally it references
						   //the m_content field (see constructor)

		//sequence of the tspan attributes. Each can be NULL, which means 
		//that it has not appear. Thses elements are stored in the main storage,
		//so they are just pointers, not released here.
		earray *m_x;
		earray *m_y;
		earray *m_dx;
		earray *m_dy;
		earray *m_rotation;
		
		unsigned m_fcount; //the size of the longest svg::earray among 
						   //m_x, m_y, m_dx, m_dy, m_rotation
	
}; //tspan

}; //namespace svg
#endif //SVG_TSPAN_INCLUDED

