//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_VERTEX_SOURCE_INCLUDED
#include "svg_basics.h"
#define SVG_VERTEX_SOURCE_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// vsource: abstract class, defining the interface required for
//			a source of the vertex commands. Prepared for the 
//	req_renderer class, where two kinds of vertex source can be used:
//
//			1. classic agg::path_base<agg::vertex_block_storage<float>> - 
//			   defined as type name path_storage. It keeps the geometrical
//			   data of all SVG path elements
//
//			2. text font outlines - obtained from the font_manager as 
//			   font_manager_type::path_adaptator_type. 
//
// To unify them (and the futire ones) you have to create an wrapper,
// which inherits from this svg::vertex_source class.
// This way it will be ready to be attached to req_render.m_vs object
// - the entry point for the vertex rendering pipeline.
//////////////////////////////////////////////////////////////////////
class vsource  
{
	public:
        virtual void rewind(unsigned path_id) = 0;
        virtual unsigned vertex(double* x, double* y) = 0;
}; //vsource

//////////////////////////////////////////////////////////////////////
// vs_path: a wrapper around the path_storage, that implements
//			the vertex_source interface (as an example of it)  
//////////////////////////////////////////////////////////////////////
class vs_path : public vsource
{
	public:
		vs_path(): m_source(NULL) {} //placeholder for an empty source (it may happen!)
		vs_path(path_storage &p): m_source(&p) {}
        virtual void rewind(unsigned path_id) {if (m_source) m_source->rewind(path_id);}
        virtual unsigned vertex(double* x, double* y) 
			{
				if(m_source) 
					return m_source->vertex(x, y);
				else
					return agg::path_cmd_stop; //something should be returned!
			}
	protected:
		path_storage *m_source;
}; //vs_path

//////////////////////////////////////////////////////////////////////
// vertex_source: an instance of this class is maintained in the
// svg::req_renderer, and can be used also in any other place, when needed. 
// It plays the role of a "switch", where you can change the vertex 
// source in the runtime, not compile time. See its source() methods.
//////////////////////////////////////////////////////////////////////
class vertex_source 
{
	public:
		vertex_source(vsource &default_source) : m_source(&default_source) {}

		//Switch into another source
		void source(vsource &vs) {m_source = &vs;}
		vsource &source() {return *m_source;}

		//functions declared as the inlines - just to not make another stack entry...
        inline void rewind(unsigned path_id) {m_source->rewind(path_id);}
        inline unsigned vertex(double* x, double* y) {return m_source->vertex(x, y);}

    private:
        vsource* m_source;
};

}; //namespace svg
#endif //SVG_VERTEX_SOURCE_INCLUDED

