//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svgtex.h"
#include "tex_texel.h"

namespace tex
{
	//Creates a texel from given pixel of the PixelFormat data
	//Arguments:
	//	pf:		the pixel buffer
	//	x,y:	position of the pixel within the pixel buffer
	//	slope:	true, when the dx,dy slope has to be calculated
	//It is the only one constructor, that can return a texel
	//with non-zero m_dx, m_dy.
	texel::texel(pixfmt &pf, int32 x, int32 y, bool slope)
	{
		//Clamp values! - pixf.pix_ptr() does not checks them.
		//and pixf has always buffer having a border, <tile_border>
		//pixels thick:
		x = CLAMP(x, -tile_border+1, (int) pf.width()+tile_border-1);
		y = CLAMP(y, -tile_border+1, (int) pf.height()+tile_border-1);

		*this = texel(pf.pix_ptr(x,y));
		if (slope)
		{
			//dx,dy: difference with the neighbours :
			int32u tr = intensity(pf.pix_ptr(x+1,y));
			int32u bl = intensity(pf.pix_ptr(x,y+1));
			int32u br = intensity(pf.pix_ptr(x+1,y+1));
		
			calc_slope(i, tr, bl, br, &dx, &dy);
		}
	}
/*----------------------------------------------------------------------------------*/	
	//Helper function, used in the filter calculations: 
	//fills four neighbour texels with data
	//Arguments:
	//	radius:	the level of the averaging: radius of the area,
	//			containing texels to be averaged. 0 returns just
	//			a texel corresponding to the pixel at x,y.
	//			It is assumed, that here <radius> is >= 1
	//	x,y:	coordinates of the source pixel in the pixel format <pf>
	//	pf:		pixel format - a buffer with the pixels
	//	p_tl:	top-left neighbour - to be filled with data
	//	p_tr:	top-right neighbour - to be filled with data
	//	p_bl:	bottom-left neighbour - to be filled with data
	//	p_br:	bottom-right neighbour - to be filled with data
	//One of the p_* texels is the texel corresponding to x,y pixel:
	//- when <radius>is odd, it is <p_tl>, when even - <p_br>
	void texel::get_neighbours(int32u radius, int32 x, int32 y, pixfmt &pf,
						texel &p_tl, texel &p_tr, texel &p_bl, texel &p_br)
	{
		radius --;
		if(radius %2) // radius is odd
		{
			p_tl	= sample(radius, x-1,	y-1, pf);
			p_tr	= sample(radius, x,		y-1, pf);
			p_bl	= sample(radius, x-1,	  y, pf);
			p_br	= sample(radius, x,		  y, pf);
		}
		else //radius is even
		{
			p_tl	= sample(radius, x,		  y, pf);
			p_tr	= sample(radius, x+1,	  y, pf);
			p_bl	= sample(radius, x,		y+1, pf);
			p_br	= sample(radius, x+1,	y+1, pf);
		}
	}
/*----------------------------------------------------------------------------------*/	
	//Helper function, used in the filter calculations.
	//Calculates a texel of u,v coordinates exactly in the center of x,y pixel
	//Arguments:
	//	radius:	the level of the averaging: radius of the area,
	//			containing texels to be averaged. 0 returns just
	//			a texel corresponding to the pixel at x,y.
	//	x,y:	coordinates of the source pixel in the pixel format <pf>
	//	pf:		pixel format - a buffer with the pixels
	texel texel::sample(int32u radius, int32 x, int32 y, pixfmt &pf)
	{
		if (radius > 0)
		{
			texel p_tl, p_tr, p_bl, p_br;
			get_neighbours(radius,x,y,pf, p_tl, p_tr, p_bl, p_br);

			return 
				texel(	//... (">> 2" is simpler than "/4"):
						(p_tl.i + p_tr.i + p_bl.i + p_br.i) >> 2, 
						(p_tl.r + p_tr.r + p_bl.r + p_br.r) >> 2, 
						(p_tl.g + p_tr.g + p_bl.g + p_br.g) >> 2, 
						(p_tl.b + p_tr.b + p_bl.b + p_br.b) >> 2, 
						(p_tl.a + p_tr.a + p_bl.a + p_br.a) >> 2,
						//(p_tl.dx + p_bl.dx) / 2, 
						//(p_tl.dy + p_tr.dy) / 2
						0,
						0
					 );
		}
		else //radius = 0, x, y already within the image size (with border!) 
		{
			return texel(pf,x,y);			
		}

	}
/*----------------------------------------------------------------------------------*/	
	//Returns a texel, which is the result of averaging of the pixels 
	//around x,y, within given radius.
	//Arguments:
	//	radius:	the level of the averaging: radius of the area,
	//			containing texels to be averaged. 0 returns just
	//			a texel corresponding to the pixel at x,y.
	//	x,y:	coordinates of the source pixel. They can be a "big"
	//			values, exceeding the <pf> dimensions. It is assummed,
	//			that the <pf> describes a fragment of this large picture
	//			build from many similar pixel formats of identical size.
	//			Another assumption - <pf> is the proper one, i.e. fragment
	//			which conatins the pixel corresponing to x,y on the large
	//			picture.
	//	pf:		pixel format - a buffer with the pixels
	//	tu,tv	pixel coordinates, corresponding to x, y. They can differ
	//			from x,y values by less than 1.0, never more!
	texel texel::sample(int32u radius, int32 x, int32 y, pixfmt &pf, float tu, float tv)
	{
		if (radius > 0)
		{
			//Values of tu and tv are between -0.5, size - 0.5.
			//Negative values produces bad ratios, so we can 
			//update them a little, to produce ptoper u_ratio or v_ratio
			//(they are not used for any other purpose here, so it is possible
			//to do such "dirty" manipulation)
			if (tu < 0.0 && x == 0) tu += 1.0;
			if (tv < 0.0 && y == 0) tv += 1.0;

			int32u u_ratio = agg::uround((tu - x) * sub_scale);
			int32u v_ratio = agg::uround((tv - y) * sub_scale);
			int32u u_opposite = sub_scale - u_ratio;
			int32u v_opposite = sub_scale - v_ratio;

			texel p_tl, p_tr, p_bl, p_br;

			x %= pf.width();
			y %= pf.height();

			get_neighbours(radius,x,y,pf, p_tl, p_tr, p_bl, p_br);
				
			
			texel result
					(
						w_avg(p_tl.i, p_tr.i, p_bl.i, p_br.i, 
											u_ratio, v_ratio, u_opposite, v_opposite),
						w_avg(p_tl.r, p_tr.r, p_bl.r, p_br.r, 
											u_ratio, v_ratio, u_opposite, v_opposite),
						w_avg(p_tl.g, p_tr.g, p_bl.g, p_br.g, 
											u_ratio, v_ratio, u_opposite, v_opposite),
						w_avg(p_tl.b, p_tr.b, p_bl.b, p_br.b, 
											u_ratio, v_ratio, u_opposite, v_opposite),
						w_avg(p_tl.a, p_tr.a, p_bl.a, p_br.a, 
											u_ratio, v_ratio, u_opposite, v_opposite),
						0, //dx and dy are calculated below:
						0
					);

			//calc dx and dy here:
			calc_slope(p_tl.i, p_tr.i, p_bl.i, p_br.i, &(result.dx), &(result.dy),
											u_opposite, u_ratio, v_opposite, v_ratio);
			return result;
		}
		else //radius = 0 - ignore required shares, but request the slope:
		{
			return texel(pf,x%pf.width(),y%pf.height(),true);			
		}
	}

/*----------------------------------------------------------------------------------*/	
	//Mixes this texel with another one, with given weight
	//Arguments:
	//	p:			another texel, which is mixed
	//	w_ratio:	share of p in the result (subscaled to integer)
	void texel::mix(texel &p, int32u w_ratio)
	{
		//int32u w_ratio = agg::uround(weight*sub_scale);
		int32u w_opposite = sub_scale - w_ratio;
		i = w_avg(i,p.i,w_ratio,w_opposite);
		r = w_avg(r,p.r,w_ratio,w_opposite);
		g = w_avg(g,p.g,w_ratio,w_opposite);
		b = w_avg(b,p.b,w_ratio,w_opposite);
		a = w_avg(a,p.a,w_ratio,w_opposite);

		//dx and dy have the sign, so we have to divide, not to shift:
		dx = (int(w_opposite) * dx + int(w_ratio) * p.dx) / sub_scale;
		dy = (int(w_opposite) * dy + int(w_ratio) * p.dy) / sub_scale;
	}

}//end namespace tex
